//File:    IDPrior.h
//Purpose: Sets the priors for the binomial model  and continuous model parameters these methods are general
//         and should be able to be used across  classes without much modification
//Creator: Matt Wheeler
//Date   : 12/18/2017
//Changes: 4/13/2018 - Changed the class IDbinomPrior to IDPrior and 
//                     added typdef for IDbinomPrior.  This allows this 
//                     basic definition to be used across types of models (continuous/dichotomous etc)
//
//
//

#pragma once
#define _USE_MATH_DEFINES
#include <Eigen/Dense>
#include <cmath>
#include <math.h>


// IDPrior
// Creates a class where each parameter specified is given an independent
// prior.  This is done in the matrix prior_spec. The first column of this
// matrix is either 1 or 2, which indicates if the parameter is normally (1)
// or log normally (2) distributed. The second column specifies the prior mean
// and the third column specifies the prior standard deviation. 
class IDPrior{
public:

	IDPrior(IDPrior &M){
		
		prior_spec = M.prior_spec; 
	}

	// Binomial Prior Coinstructor
	IDPrior(Eigen::MatrixXd tP):prior_spec(tP){
		// TODO: do some error checking to make sure you are passing in
		// the correct size matrix and that the bounds make sense in relation
		// to the mean. 
	};
	
	void set_prior(Eigen::MatrixXd tP){
		prior_spec = tP; 
	}
	double neg_log_prior(Eigen::MatrixXd theta);
	Eigen::MatrixXd prior_mean(); 

	Eigen::MatrixXd lowerBounds() {
		return prior_spec.col(3);
	}

	Eigen::MatrixXd upperBounds() {
		return prior_spec.col(4);
	}

private:
	// Matrix that defines the prior specifications 
	// for all of the parameters in theta
	// First Column - Type of Prior
	// Second Column - mean
	// Third Column  - Dispersion
	// Fourth Column - Lower Bound
	// Fith Column   - Upper Bound
	Eigen::MatrixXd prior_spec;
};

//set different typdefs for the type of prior continuous or binomial
//model
typedef IDPrior IDbinomPrior ; 
typedef IDPrior IDcontinuousPrior ; 


///////////////////////////////////////////////////////////
// Function: neg_log_prior(Eigen::MatrixXd theta)
// Return the negative log prior for the prior specification
// Input: theta a kx1 matrix
// Output: Returns a real value based upon the prior for the given 
//         paramer defined in prior_spec. Here prior_spec is a kx3
//         matrix.  The first column defines the type of prior 1-normal
//         2-log-normal, 0 improper.  The second colunn defines the mean. 
//         For the improper prior, this is not used but it still defines an \
	//         option for the initialization.  The third column specifies the 
//		   dispersion parameter. 
double IDPrior::neg_log_prior(Eigen::MatrixXd theta) {
	double returnV = double(theta.rows())*log(0.5*M_2_SQRTPI * M_SQRT1_2);
	double mean = 0;
	double sd = 0;
	// loop over the prior specification in prior_spec
	// when  it is 1 - Normal Prior
	// when  it is 2 - Log normal prior.
	for (int i = 0; i < theta.rows(); i++) {
		int t = int(prior_spec(i, 0));
		switch (t) {
		case 1:
			mean = (theta(i, 0) - prior_spec(i, 1));
			sd = prior_spec(i, 2);
			returnV += -log(sd) - 0.5*mean*mean / (sd*sd);
			break;
		case 2:
			mean = (log(theta(i, 0)) - prior_spec(i, 1));
			sd = prior_spec(i, 2);
			returnV += -log(sd) - log(theta(i, 0)) - 0.5*mean*mean / (sd*sd);
			break;
		default: // in the default case we remove all prior info
			returnV -= log(0.5*M_2_SQRTPI * M_SQRT1_2);
			break;
		}

	}
	return -1.0*returnV;
};
//////////////////////////////////////////////////////
// Function: prior_mean()
// Purpose returns the prior_means
// Output: Returns the prior mean for normal and median for the log normal distributions
//		   for improper prior it returns the value that is in column 2, which 
//         can be used for initial starting points. 
Eigen::MatrixXd IDPrior::prior_mean() {
	Eigen::MatrixXd pmean(prior_spec.rows(), 1);

	for (int i = 0; i < prior_spec.rows(); i++) {
		int t = int(prior_spec(i, 0));
		switch (t) {
		case 2:
			pmean(i,0) = exp(prior_spec(i, 1));			
			break;
		default: // in the default case we remove all prior info
			pmean(i, 0) = prior_spec(i, 1);
			break;

		}
	}
	return pmean;
};
