#pragma once


#if defined WIN32 || defined _WINDOWS

#  ifdef BMDS_MODELS_EXPORTS
#    define BMDS_ENTRY_API __declspec(dllexport)
#  else
#    define BMDS_ENTRY_API __declspec(dllimport)
#  endif // BMDS_MODELS_EXPORTS

typedef BSTR BMDS_MODEL_ID;
#else
typedef char * BMDS_MODEL_ID;
#endif // WIN32

#define BMDS_BLANK_VALUE -9999

static const int NUM_PRIOR_COLS = 5;

// The "pack(4)" pragma is required because Excel requires four-byte aligned structures,
// but Visual Studio uses eight-byte boundaries by default.

#pragma pack(4)

enum DModelID_t {
  eDHill = 1,
  eGamma = 2,
  eLogistic = 3,
  eLogLogistic = 4,
  eLogProbit = 5,
  eMultistage = 6,
  eProbit = 7,
  eQLinear = 8,
  eWeibull = 9
};

enum BMDSPrior_t {
  eNone = 0,
  eNormal = 1,
  eLognormal = 2
};

enum BMRType_t {
  eAbsoluteDev = 1,
  eStandardDev = 2,
  eRelativeDev = 3,
  ePointEstimate = 4,
  eExtra = 5, // Not used
  eHybrid_Extra = 6,
  eHybrid_Added = 7
};

enum RiskType_t {
  eExtraRisk = 1,
  eAddedRisk = 2
};

enum VarType_t {
  eVarTypeNone = 0,
  eConstant = 1,
  eModeled = 2
};

struct BMDSOptions_t {
  int bmrType;
  double bmr;
  double alpha;
  double background;
  double tailProb;  // Valid only for hybrid bmr type
  bool bLognormal;  // Valid only for continuous models
  int degree; // Valid for polynomial type models; for exponential, identifies the submodel
  VarType_t varType;
};

struct BMDS_D_Opts1_t {
  double bmr;
  double alpha;
  double background;
};

struct BMDS_D_Opts2_t {
  int bmrType;
  int degree; // Polynomial degree for the multistage model
};

struct gofRow {
  double dose;
  double estProb;  // Model-estimated probability for dose
  double expected; // Expected dose-response according to the model
  double observed;
  double size;
  double scaledResidual;
};
typedef struct gofRow GoFRow_t;

struct dGoF {
  double chiSquare;
  double pvalue;
  GoFRow_t *pzRow;
  int df;
  int n; // Number of rows
};
typedef struct dGoF dGoF_t;

struct BMD_ANAL {
  BMDS_MODEL_ID model_id; 
	double MAP; // Equals the -LL for frequentist runs
	double BMD; 
	double BMDL;
	double BMDU;
    double AIC;
    double BIC_Equiv; // BIC equivalent for Bayesian runs
	double *PARMS;
    struct DichotomousDeviance_t *deviance;
    dGoF_t *gof; // Goodness of Fit
    bool *boundedParms;
	int  nparms; 
};

struct PRIOR {
  double type; // 0= None (frequentist), 1=  normal (Bayesian), 2= log-normal (Bayesian)
  double initalValue;
  double stdDev; // Only used for type= 1 or 2
  double minValue;
  double maxValue;
};

struct PARM_PRIORS {
  int nparms;
  PRIOR *priors;
};

struct MA_PRIORS {
	double *mean_logistic; double *sd_logistic; 
	double *mean_probit;   double *sd_probit  ;
	double *mean_loglogit; double *sd_loglogit; 
	double *mean_logprobit; double *sd_logprobit; 
	double *mean_weibull;  double *sd_weibull; 
	double *mean_gamma;    double *sd_gamma; 
	double *mean_mult2;    double *sd_mult2;
	double *mean_qlinear;  double *sd_qlinear; 
	double *mean_hill;     double *sd_hill; 
};

struct MA_ANALYSIS {
	double BMD; double BMDL; double BMDU; 
	double *bmd;
	double *bmdl; 
	double *bmdu; 
	double *post_probs; 
};

enum BMDSInputType_t {
  unused = 0,
  eCont_2 = 1, // Summarized dose-responses
  eCont_4 = 2, // Individual dose-responses
  eDich_3 = 3, // Regular dichotomous dose-responses
  eDich_4 = 4  // Dichotomous d-r with covariate (e.g., nested)
};

struct BMDSInputData_t {
  double dose;
  double response; // Mean value for summary data
  double groupSize;
  double col4; // stddev for cont_4 or covariate for dich_4
};

struct DichotomousDeviance_t {
  double llFull; // Full model log-likelihood
  double llReduced; // Reduced model log-likelihood
  double devFit; // Fit model deviance
  double devReduced; // Reduced model deviance
  double pvFit; // Fit model p-value
  double pvReduced; // Reduced model p-value
  int nparmFull;
  int nparmFit;
  int dfFit;
  int nparmReduced;
  int dfReduced;
};

#pragma pack()

extern "C" {

  BMDS_ENTRY_API
    void _stdcall bmd_MA(BMDSInputType_t inputType, BMDSInputData_t *dataIn,
      MA_PRIORS *priors, double *p_m_probs,
      BMDS_D_Opts1_t *opt1, BMDS_D_Opts2_t *opt2, int n, double * post_p,
      double *ma_bmd, double *bmd, double *bmdl, double *bmdu);

  BMDS_ENTRY_API
    int _stdcall run_dmodel2(DModelID_t *p_m_id, BMD_ANAL *returnV,
                            BMDSInputType_t *p_inputType, BMDSInputData_t *dataIn,
                            PRIOR *priorsIn, BMDS_D_Opts1_t *opt1, BMDS_D_Opts2_t *opt2, int *p_n);

  BMDS_ENTRY_API
    int _stdcall run_dmodel(DModelID_t *p_m_id, BMD_ANAL *returnV,
                          BMDSInputType_t *p_inputType, BMDSInputData_t *dataIn,
                          PRIOR *priorsIn, BMDSOptions_t *options, int *p_n);

  // ***** DEPRECATED entry points for the individual models ***** //

  BMDS_ENTRY_API
    void _stdcall bmd_log_logistic(BMD_ANAL *returnV, BMDSInputType_t inputType,
      BMDSInputData_t *dataIn, PRIOR *priorsIn, BMDSOptions_t *options, int n);

  BMDS_ENTRY_API
    void _stdcall bmd_log_probit(BMD_ANAL *returnV, BMDSInputType_t inputType,
      BMDSInputData_t *dataIn, PRIOR *priorsIn, BMDSOptions_t *options, int n);

  BMDS_ENTRY_API
    void _stdcall bmd_weibull(BMD_ANAL *returnV, BMDSInputType_t inputType,
      BMDSInputData_t *dataIn, PRIOR *priorsIn, BMDSOptions_t *options, int n);

  BMDS_ENTRY_API
    void _stdcall bmd_gamma(BMD_ANAL *returnV, BMDSInputType_t inputType,
      BMDSInputData_t *dataIn, PRIOR *priorsIn, BMDSOptions_t *options, int n);

  BMDS_ENTRY_API
    void _stdcall bmd_multistage(BMD_ANAL *returnV, BMDSInputType_t inputType,
      BMDSInputData_t *dataIn, PRIOR *priorsIn, BMDSOptions_t *options, int n);

  BMDS_ENTRY_API
    void _stdcall bmd_probit(BMD_ANAL *returnV, BMDSInputType_t inputType,
      BMDSInputData_t *dataIn, PRIOR *priorsIn, BMDSOptions_t *options, int n);

  BMDS_ENTRY_API
    void _stdcall bmd_logit(BMD_ANAL *returnV, BMDSInputType_t inputType,
      BMDSInputData_t *dataIn, PRIOR *priorsIn, BMDSOptions_t *options, int n);

  BMDS_ENTRY_API
    void _stdcall bmd_qlinear(BMD_ANAL *returnV, BMDSInputType_t inputType,
      BMDSInputData_t *dataIn, PRIOR *priorsIn, BMDSOptions_t *options, int n);

  BMDS_ENTRY_API
    void _stdcall bmd_hill(BMD_ANAL *returnV, BMDSInputType_t inputType,
      BMDSInputData_t *dataIn, PRIOR *priorsIn, BMDSOptions_t *options, int n);
}
