//File: statmodel.h
//Purpose: Create a basic statistical model class
//		   that is used for any general analysis, and can be 
//         rather flexible. 
//Creator: Matt Wheeler
//Date   : 12/18/2017
//Changes: 
//  Matt Wheeler
//  Date 11/06/2018 Added returnX() function to the statModel class
//
//
#pragma once
#include "stdafx.h"
#include <cmath>
#include <Eigen/Dense>
#include <nlopt.hpp>
#include "log_likelihoods.h"
#include "IDPrior.h"
#include "binomModels.h"

#include <iostream> 

using namespace std; 

struct optimizationResult {
	nlopt::result result; 
	double functionV;
	Eigen::MatrixXd max_parms; 
};

template <class LL, class PR> 
class statModel {
public: 
	statModel(LL t_L, PR t_PR,
			  std::vector<bool> b_fixed, 
			  std::vector<double> d_fixed) :log_likelihood(t_L), prior_model(t_PR),
										    isFixed(b_fixed), fixedV(d_fixed){

		if (isFixed.size() != fixedV.size()) {
			throw std::runtime_error(std::string("Statistical Model: Fixed parameter constraints are same size"));
		}

		if (isFixed.size() != log_likelihood.nParms()) {
			throw std::runtime_error(std::string("Statistical Model: Fixed number of parameter constraints not equal to number of parameters in likelihood model."));
		}
	}; 

	/////////////////////////////////////////////////////////////////////
	//compute the variance covariance matrix as well as the 
	//gradient using finite derivatives. 
	Eigen::MatrixXd varMatrix(Eigen::MatrixXd theta);
	Eigen::MatrixXd gradient(Eigen::MatrixXd v);
	//////////////////////////////////////////////////////////////////////

	

	double negPenLike(Eigen::MatrixXd x) {
		///////////////////////////////////////////
		for (int i = 0; i < isFixed.size(); i++) {
			if (isFixed[i]) {
				x(i, 0) = fixedV[i]; 
			}
		}
		///////////////////////////////////////////
		double a = log_likelihood.negLogLikelihood(x);
		double b = prior_model.neg_log_prior(x); 
		return a + b; // log_likelihood.negLogLikelihood(x) + prior_model.neg_log_prior(x);
	};

	int nParms() {
		return log_likelihood.nParms(); 
	};

	Eigen::MatrixXd startValue() {
		return prior_model.prior_mean();
	}

	Eigen::MatrixXd parmLB() {
		return prior_model.lowerBounds(); 
	}

	Eigen::MatrixXd parmUB() {
		return prior_model.upperBounds();
	}
	// functions to get and set the values of the central estimate of the 
	// statistical model
	virtual void setEST(const Eigen::MatrixXd x) { 
		Eigen::MatrixXd thetaX = x;
		// worry about the fixed parameters
		for (int i = 0; i < isFixed.size(); i++) {
			if (isFixed[i]) {
				thetaX(i, 0) = fixedV[i];
			}
		}
		theta = thetaX; 
	};

	virtual Eigen::MatrixXd getEST() {
		Eigen::MatrixXd thetaX = theta;
		// worry about the fixed parameters
		for (int i = 0; i < isFixed.size(); i++) {
			if (isFixed[i]) {
				thetaX(i, 0) = fixedV[i];
			}
		}
		return thetaX; 
	};
	
	Eigen::MatrixXd returnX(){
			return log_likelihood.returnX(); 
	}
public:
	// A stat model has a Log Likelihood and 
	// A prior model over the parameters
	// Note: The prior over the parameters does not have to be an actual 
	// prior. It can merely place bounds (e.g., box, equality) on functions
	// of the parameters. 
	 LL log_likelihood;
	 PR prior_model; 
	// The parameter vector that the log_likelihood and the prior model
	// both depend upon
	 std::vector<bool>    isFixed; 
	 std::vector<double>  fixedV; 
	 Eigen::MatrixXd theta; 

};

/////////////////////////////////////////////////////////////////
// Function: gradient(Eigen::MatrixXd v)
// Purpose : Performs a finite difference evaluation of the
//           gradient 
// Input   :
//          Eigen::MatrixXd v - vector of parameters to model
// 
// Output  :
//          double * g      - vector of gradients cooresponding to the
//                            parameters v
///////////////////////////////////////////////////////////////////
template <class LL, class PR>
Eigen::MatrixXd statModel<LL,PR>::gradient(Eigen::MatrixXd v) {

	Eigen::VectorXd h(log_likelihood.nParms());
	double mpres = pow(1.0e-16, 0.5);
	double x, temp;
	double derEst;
	double f1 = 0.0; double f2 = 0.0;
	Eigen::MatrixXd  hvector=v;
	Eigen::MatrixXd  g(log_likelihood.nParms(), 1);

	for (int i = 0; i < log_likelihood.nParms(); i++)
	{
		x = v(i, 0);
		if (abs(x) > DBL_EPSILON) {
			h[i] = mpres * (abs(x));
			temp = x + h[i];
			h[i] = temp - x;
		}else {
			h[i] = mpres;
		}
	
	}
	/*find the gradients for each of the variables in the likelihood*/
	for (int i = 0; i < log_likelihood.nParms(); i++) {
		/*perform a finite difference calculation on the specific derivative*/
		x = v(i,0);
		// add h
		hvector(i,0) = x + h[i];
		f1 = negPenLike(hvector);
		// subtract h
		hvector(i,0) = x - h[i];
		f2 = negPenLike(hvector);
		// estimate the derivative
		g(i,0) = (f1 - f2) / (2.0*h[i]);
		hvector(i,0) = x;
	}
	return g; 
}

////////////////////////////////////////////////////////////////////
// Function: varMatrix(Eigen::MatrixXd theta)
// Purpose:  Computes the inverse hessian of the penalized likelihood 
//			 given the parameter theta. 
// Input  :  Eigen::MatrixXd theta kx1 parameter value equal to the 
//           number of parameters in the "statistical model"
// Output :  Eigen::MatrixXd k x k symmetric positive definite matrix 
//			 Inv Hessian of the penalized likelihood. 
/////////////////////////////////////////////////////////////////////
template <class LL, class PR> 
Eigen::MatrixXd statModel<LL,PR>::varMatrix(Eigen::MatrixXd theta) {

	Eigen::MatrixXd m(log_likelihood.nParms(), log_likelihood.nParms());
	double temp;
	double h = 1e-8;
	double mpres = pow(1.0e-16, 0.333333);

	Eigen::MatrixXd tempVect(log_likelihood.nParms(), 1);
	double hi, hj;
	
	for (unsigned int i = 0; i < log_likelihood.nParms(); i++) {
		for (unsigned int j = 0; j < log_likelihood.nParms(); j++) {
			
			double x = theta(i, 0);  //compute the hessian for parameter i-j
			if (abs(x) > DBL_EPSILON) {
				h = mpres * (abs(x));
				temp = x + h;
				hi = temp - x;
			}else {
				hi = mpres;
			}

			x = theta(j, 0);
			if (abs(x) > DBL_EPSILON) {
				h = mpres * (abs(x));
				temp = x + h;
				hj = temp - x;
			}else {
				hj = mpres;
			}

			temp = 0.0;

			if (i == j) {
				// on diagonal variance estimate 
				// eg the 2nd partial derivative
				tempVect = theta;
				tempVect(i, 0) = tempVect(i, 0) + 2 * hi;
				temp += -1.0*negPenLike(tempVect);
				tempVect = theta;
				tempVect(i, 0) = tempVect(i, 0) + hi;
				temp += 16.0*negPenLike(tempVect);
				temp += -30.0*negPenLike(theta);
				tempVect = theta;
				tempVect(i, 0) = tempVect(i, 0) - hi;
				temp += 16.0*negPenLike(tempVect);

				tempVect = theta;
				tempVect(i, 0) = tempVect(i, 0) - 2 * hi;
				temp += -1.0*negPenLike(tempVect);
				temp = temp / (12 * hi*hi);
				m(i, i) = temp;

			}else {

				// off diagonal variance estimate 
				// eg the first partial of x wrt the
				// first partial  of y

				tempVect = theta;
				tempVect(i, 0) = tempVect(i, 0) + hi; tempVect(j, 0) = tempVect(j, 0) + hj;
				temp += negPenLike(tempVect);
				
				tempVect = theta;
				tempVect(i, 0) = tempVect(i, 0) + hi; tempVect(j, 0) = tempVect(j, 0) - hj;
				temp += -1.0*negPenLike(tempVect);

				tempVect = theta;
				tempVect(i, 0) = tempVect(i, 0) - hi; tempVect(j, 0) = tempVect(j, 0) + hj;
				temp += -1.0*negPenLike(tempVect);

				
				tempVect = theta;
				tempVect(i, 0) = tempVect(i, 0) - hi;
				tempVect(j, 0) = tempVect(j, 0) - hj;
				temp += negPenLike(tempVect);
				temp = temp / (4 * hi*hj);
				m(i, j) = temp;

			}

		}

	}
	// m = m.inverse();
	
	return m.inverse();
}

/////////////////////////////////////////////////////////////////
//Function: neg_pen_likelihood(unsigned n,
//								const double *b,
//								double *grad,
//								void *data)
//Purpose: Used to estimate the negative penalized likelihood from the statMod class
//         in the proper NLOPT format
//Input: 
//         unsigned n: The length of the vector of parameters to the optimizer
//         const double *b   : A pointer to the place in memory the vector resides
//         double *gradient  : The gradient of the function at b, NULL if unused. 
//                             It is currently unused. 
//         void    *data     : Extra data needed. In this case, it is a statModel<LL,PR> object,
//							   which is used to compute the negative penalized likelihood
//////////////////////////////////////////////////////////////////
template <class LL, class PR> 
double neg_pen_likelihood(unsigned n,
						  const double *b,
						  double *grad,
						  void *data) {
	statModel<LL, PR> *model = (statModel<LL, PR>*) data; 
	Eigen::MatrixXd theta(n, 1);
	for (int i = 0; i < n; i++) theta(i, 0) = b[i];

	// if an optimizer is called that needs a gradient we calculate
	// it using the finite difference gradient
	if (grad) {
		Eigen::MatrixXd mgrad = model->gradient(theta);
		for (int i = 0; i < model->nParms(); i++) {
				grad[i] = mgrad(i, 0);
		
		}
	}
	 
	return model->negPenLike(theta); 
}

////////////////////////////////////////////////////////////
//Function: startValue_F(statModel<LL, PR>  *M) 
//Purpose : Given a statistical model M it produces a starting 
//          value based upon one iteration of a Fisher scoring algorithm.
//          As there is no reason to assume the single iteration will respect
//			the parameter bounds, the routine checks if the parameter bounds
//		    are violated; if any are, it moves in the direction of the change
//          in half power increments until the bounds are not violated. This value 
//	        is returned. 
//Input    : statModel<LL,PR> *M - Pointer to a statistical value whose central
//			 tendency values in PR respect the constraints of the model.
//Output   : A starting value to the optimizer that *should* allow the optimizer to
//         : proceed quickly to the optimum. 
/////////////////////////////////////////////////////////////
template <class LL, class PR>
std::vector<double> startValue_F(statModel<LL, PR>  *M,
								 std::vector<double> lb, 
							     std::vector<double> ub)
{
	std::vector<double> x(M->nParms());
	bool correctBounds = false;
	double multiplier = 1.0;
	int j = 0;
	Eigen::MatrixXd start = M->startValue();
	Eigen::MatrixXd test;
	 
	Eigen::MatrixXd FS = M->varMatrix(start)*(M->gradient(start));


	for (int ii = 0; ii < 3; ii++) {
		j = 0; 
		multiplier = 1.0; 
		correctBounds = false; 
		while (!correctBounds && j < 16) {
			// Do a Fisher scoring step to develop starting values for the optimizer.
			// The multiplier is in [0,1], and it tells us how far to go.
			// FS is the full step distance suggested by the Fisher Scoring Algorithm
			// Note: We subtract vs add because we are dealing with the negative penalized
			// log likelihood, and we need to correct for that.  
			test = start - multiplier * FS;

			// Test that the new start value maintains the constraints
			correctBounds = true;
			for (int i = 0; i < M->nParms(); i++) {
				if (lb[i] > test(i, 0) || ub[i] < test(i, 0)) {
					correctBounds = false;
					break;
				}
			}
			multiplier *= 0.5; j++;

		}
		if (correctBounds) { // if we have found a new
							// value with the correct bounds
							// try to get closer : do this three times
			start = test; 
			FS = M->varMatrix(start)*(M->gradient(start));
			
		}
		
	}
	// If we still havn't hit the right bounds or there
	// is something screwey in our estimate
	// the starting point is the initial start.
	int NAN_COUNT = 0; double kkk = 0.0; 
		for (int ii = 0; ii < test.rows(); ii++) {
			
			if (isnan(abs(test(ii, 0)))) NAN_COUNT++;
		}

	if (!correctBounds || NAN_COUNT )
		test = M->startValue();
	// Copy the answer into x and return. 
	
	if (!(M->negPenLike(M->startValue()) > M->negPenLike(test))) {
		test = M->startValue();
	}
test = M->startValue(); //***///****//******
	for (int i = 0; i < M->nParms(); i++)	x[i] = test(i, 0);
	return x; 

}

////////////////////////////////////////////////////////////////////
//Function: findMAP(statMod<LL,PR> *M)
//Find the maximum a posteriori estimate given a statistical 
//model
//Input : statMod<LL,PR> *M - A given statistical model; 
//Output: statMod<LL,PR> *M - The model with it's MAP parameter set.
template <class LL, class PR> 
optimizationResult findMAP(statModel<LL, PR>  *M) {
	optimizationResult oR; 

	Eigen::MatrixXd temp_data = M->parmLB(); 
	std::vector<double> lb(M->nParms()); 
  for (int i = 0; i < M->nParms(); i++) lb[i] = temp_data(i, 0);
	temp_data = M->parmUB(); 
	//cout << temp_data << endl; 
	std::vector<double> ub(M->nParms()); 
	for (int i = 0; i < M->nParms(); i++) ub[i] = temp_data(i, 0);

	std::vector<double> x =  startValue_F(M, lb, ub);
//	for (int i = 0; i < x.size(); i++) cout << x[i] << endl; 

	double minf;
	nlopt::result result = nlopt::FAILURE;

	std::vector<double> init(x.size());
	//for (int i = 0; i < x.size(); i++) init[i] = 1e-4;
	// set up a bunch of differnt plausible optimizers in case of failure
	// the first one is mainly to get a better idea of a starting value  
	// though it often converges to the optimum. 
	nlopt::opt opt1(nlopt::LN_SBPLX, M->nParms());
	nlopt::opt opt2(nlopt::LN_BOBYQA, M->nParms());
	nlopt::opt opt3(nlopt::LD_LBFGS,M->nParms());
	nlopt::opt opt4(nlopt::LN_COBYLA,M->nParms()); 
    nlopt::opt opt5(nlopt::LD_SLSQP,M->nParms());

	nlopt::opt *opt_ptr; 
	
	
	// look at 5 optimization algorithms :-)
	for (int opt_iter = 0; opt_iter < 5; opt_iter++){
		
	
		switch(opt_iter){		
			case  0:
				opt_ptr = &opt1; 
				opt_ptr->set_maxeval(20000);
				break; 
			case  1: 
				opt_ptr = &opt2 ;
				opt_ptr->set_maxeval(50000);
				break; 
			case 2:
				opt_ptr = &opt3;
				opt_ptr->set_maxeval(50000);               
				break; 
			case 3:
				opt_ptr = &opt4; 
				opt_ptr->set_maxeval(50000);               
			default : 
				opt_ptr = &opt5; 
				opt_ptr->set_maxeval(10000);
				break; 				
		}
	
		opt_ptr->set_lower_bounds(lb); 
		opt_ptr->set_upper_bounds(ub);
		opt_ptr->set_ftol_rel(1e-8);
		
		opt_ptr->set_min_objective(neg_pen_likelihood<LL,PR>, M);

		////////////////////////////////////////////////
		//////////////////////////////////////////////
		// set the start distance to a size
		// nlopt's default options suck
		//opt_ptr->set_initial_step(1e-4);		
		///////////////////////////////////////////////

		try{
		 result = opt_ptr->optimize(x, minf);
		 // note even if it doesn't converge the x will be updated
		 // to the best value the optimizer ever had, which will allow
		 // the next optimizer to carry on.
		 if (opt_iter > 1 
			&& result > 0 
			&& result < 5) {
		   //cout << opt_iter; 
			opt_iter = 10;  // if it made it here it will break the loop	    
		 }

		 
		}
    catch (const std::invalid_argument &exc) {
			
			cout << opt_iter << "  SHIT!!" << endl;
    }
    catch (nlopt::roundoff_limited) {
			
			//cout << opt_iter << "  RL SHIT!!" << endl;
		}
		catch (nlopt::forced_stop) {
			
		  //cout << opt_iter << "  FS SHIT!!" << endl;
		}
		catch (const std::exception &exc) {
		  //cout << opt_iter << "Here!!" << endl; 

		}
	}
	// copy the result into a temporary vector
	Eigen::Map<Eigen::MatrixXd> d(x.data(),M->nParms(), 1);
	oR.result = result; 
	oR.functionV = minf; 
	oR.max_parms = d; 
	//cout << d << endl;
	// TODO: do something if the optimization fails
  M->setEST(d);
  if (result < 0) {
    //cout << "something freakin' failed" << endl;
    cerr << __FUNCTION__ << " at line: " << __LINE__ << " result= " << result << endl;
    //ofstream file;
    //file.open("bmds.log", fstream::app);
    //file << __FUNCTION__ << " at line: " << __LINE__ << " result= " << result << endl;
    //file.close();
  }

	return oR;
}

