#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* This is copied from benchmark.h so we don't need to include it */
/*data type for ANOVA summary*/              
typedef struct ana_struct {
	double SS;
	double MSE;
	int    DF;
	double TEST;
} AnaList;

#include "ms_combo_dll.h"

extern FILE *fp_out;

void myPrintText(FILE *fp, const char *pText)
{
  fprintf(fp, "%s\n", pText);
  fflush(fp);
  return;
}

void myPrintInitParms(FILE *fp, int nparm, int Spec[], double Parms[], char *tparms[])
{
  int i;
  for (i = 1; i <= nparm; i++) {
    if (Spec[i] == 0) fprintf (fp, "%31s = %12.6g\n" , tparms[i - 1], Parms[i]);
    else fprintf (fp, "%31s = %12.6g   Specified\n" , tparms[i - 1], Parms[i]);
  } /* end for */
  fflush(fp);
  return;
} // printInitParms()

void myErrorPrt(FILE *fp, const char *text)
{
  fprintf(stderr, "\n%s\n", text);
  fprintf(fp, "\n%s", text);
  exit (1);                                
}

void myWarning(FILE *fp, const char *text)
{
  fprintf(fp, "\n%s\n", text);
  return;
}

/******************************************************
**  OUTPUT_DTMS3ANOVA--output ANOVA table values for a
*                      3-parameter dichotomous model.
*******************************************************/
void myOUTPUT_DTMS3PARMS (int nparm, int Spec[], int bounded[], double Parms[],
			  char *tparms[], double **vcv, int print_SE)
{
  int i, jvar;

  /*output to *.out file */
  myPrintText(fp_out, "\n\n\n                                Parameter Estimates");
  myPrintText(fp_out,     "\n       Variable           Estimate             Std. Err. ");
  jvar = 0;
  for (i = 1; i <= nparm; i++) {
    if(print_SE > 0) {
      if (Spec[i] == 0) {
	if (bounded[i] == 0) {
	  jvar++;
	  if (vcv[jvar][jvar] > 0) {
	    fprintf (fp_out, 
		     "%15s %19.6g %19.6g\n"
		     , tparms[i - 1], Parms[i],
		     sqrt (vcv[jvar][jvar]));
	  }
	  else {
	    fprintf (fp_out, 
		     "%15s %19.6g %19s\n"
		     , tparms[i - 1], Parms[i],
		     "NA");
	  }
	}
	else {
	  fprintf (fp_out, 
		   "%15s %19.6g             Bounded\n"
		   , tparms[i - 1], Parms[i]);
	}
      }
    }
    else { //print_SE == 0
      if (Spec[i] == 0) {
	if (bounded[i] == 0) {
	  jvar++;
	  fprintf (fp_out, 
		   "%15s %19.6g %13s\n"
		   , tparms[i - 1], Parms[i],
		   "*");
	}
	else {
	  fprintf (fp_out, 
		   "%15s %19.6g             *\n"
		   , tparms[i - 1], Parms[i]);
	}
      }
    }
  } //end of for(i = 1; i <= nparm; i++)
  if(print_SE == 0) /* Don't Print Standard Error */
    fprintf (fp_out, "\n* - Indicates that this value is not calculated.\n");
} // end myOUTPUT_DTMS3PARMS()

/*********************************************************************************
 *
 *  OUTPUT_DTMS3VCV--This function will take the matrix of Likelihood
 *		    second derivatives and remove the columns/rows
 *		    that correspond to any bounded or fixed parameters,
 *		    invert this modified matrix, then output it.
 *
 *  INPUT:
 *      nparm       number of parameters in the model
 *      Spec[i]     0 if parameter i is not specified by user
 *		   1 if parameter i is specified by user
 *      parmtxt[i]  the ith parameters name
 *      vcv[][]     contains the second partial derivatives of
 *	           the likelihood function.  Element ij is
 *		   the second partial with respect to
 *		   parameters i and j
 *      bounded[i]  0 if the parameter has not been estimated
 *		   at a boundary point
 *		   1 if the parameter has been estimated at a
 *		   boundary point
 *  INPUT/OUTPUT:
 *	vcv_adj[][]-	Upon entry, is just an uninitialized matrix
 *			of size MxM where M = # of non-bounded
 *			parameters.  At return, it is the corr.
 *			matrix with all bounded and specified
 *			parameters removed
 **********************************************************************************/
void myGet_and_OUTPUT_DTMSVCV (int nparm, int Spec[], char *parmtxt[],
			       double **vcv, double **vcv_adj, int *bounded)
{
  int adj_vcv_rows, i, j, checkvar, counter, *notbounded, nboundparm;
  double cor, **temp_vcv;
  double **DMATRIX(); void FREE_DMATRIX ();
  int *IVECTOR(); void  FREE_IVECTOR ();
  int Take_Out_Bounded_Parms(int, int *, double **);
  int INVMAT(double **, int);
  
  /* Create an adjusted variance-covariance matrix that does not have
     correlations/variances for parameters that are estimated at a boundary point */

  temp_vcv = DMATRIX (1, nparm, 1, nparm);

  for (i = 1; i <= nparm; i++) {
    for (j = 1; j <= nparm; j++) {
      temp_vcv[i][j] = vcv[i][j];
    }
  }


  nboundparm = 0;
  for (i = 1; i <= nparm; i++) {
    if (bounded[i] == 1) {
      nboundparm++;
    }			/* end if */
  }				/* end for */

  notbounded = IVECTOR (1, nparm - nboundparm);
  counter = 0;

  for (i = 1; i <= nparm; i++) {
    if (bounded[i] == 0) {
      /* notbounded[j] contains the subscript in bounded that contains the
	 jth non-bounded parameter */

      counter++;
      notbounded[counter] = i;
    }			/* end if */
  }				/* end for */

  /* Adjust matrix to only contain non-fixed or bounded parameters */
  adj_vcv_rows = Take_Out_Bounded_Parms (nparm, bounded, vcv);

  for (i = 1; i <= adj_vcv_rows; i++) {
    for (j = 1; j <= adj_vcv_rows; j++) {
      vcv_adj[i][j] = vcv[i][j];
    }			/* end for j */
  }				/* end for i */


  /* Invert to get the covariance matrix */

  INVMAT (vcv_adj, adj_vcv_rows);

  /*output to screen and bmdswrk.002 temp file */
  myPrintText(fp_out, "\n\n           Asymptotic Correlation Matrix of Parameter Estimates\n");

  if (nboundparm > 0) {
    /* printf ("           ( *** The model parameter(s) "); */
    fprintf (fp_out, "           ( *** The model parameter(s) ");
    for (i = 1; i <= nparm; i++) {
      if (bounded[i] == 1) {
	/* printf (" -%s   ", parmtxt[i - 1]); */
	fprintf (fp_out, " -%s   ", parmtxt[i - 1]);
      }
    }
    fprintf (fp_out, "\n                 have been estimated at a boundary point, or have been specified by the user,");
    fprintf (fp_out, "\n                 and do not appear in the correlation matrix )\n\n");
  }

  fprintf (fp_out, "          ");

  counter = 0;

  for (i = 1; i <= nparm; i++) {
    if (bounded[i] == 0) {
      /* printf ("%15s", parmtxt[i - 1]); */
      fprintf (fp_out, "%13s", parmtxt[i - 1]);
    }
  }
  fprintf (fp_out, "\n");


  for (i = 1; i <= adj_vcv_rows; i++) {
    fprintf (fp_out, "\n%10s", parmtxt[notbounded[i] - 1]);
    for (j = 1; j <= adj_vcv_rows; j++) {
      if (vcv_adj[i][i] <= 0.0 || vcv_adj[j][j] <= 0.0) {
	fprintf (fp_out, "      NA       ");
	checkvar = 1;
      }
      else {
	cor = vcv_adj[i][j] /
	  (sqrt (vcv_adj[i][i]) * sqrt (vcv_adj[j][j]));
	fprintf (fp_out, 
		 "%13.2g", 
		 cor);
	checkvar = 0;
      }
    }
    fprintf (fp_out, "\n");
  }

  if (checkvar == 1) {
    fprintf (fp_out,
	     "\n\nNA - This parameter's variance has been estimated as zero or less.\nTHE MODEL HAS PROBABLY NOT CONVERGED!!!\n");
  }

  for (i = 1; i <= nparm; i++) {
    for (j = 1; j <= nparm; j++) {
      vcv[i][j] = temp_vcv[i][j];
    }
  }
    
  FREE_IVECTOR(notbounded,1,nparm - nboundparm);
  FREE_DMATRIX(temp_vcv,1,nparm,1,nparm);
} // myGet_and_OUTPUT_DTMSVCV()

void myOUTPUT_DTMS3ANOVA (char *anatxt[], AnaList anasum[])

{
  int Nparms;

  /*output to screen */
  myPrintText(fp_out, "\n\n\n                        Analysis of Deviance Table");
  myPrintText(fp_out, "\n       Model      Log(likelihood)  # Param's  Deviance  Test d.f.   P-value");

  /*output to *.out file */
  Nparms = anasum[3].DF + 1; /* for full model == # dose groups */
  fprintf (fp_out, 
	   "%15s %15.6g %9d\n"
	   , anatxt[0], anasum[1].SS, Nparms);
  Nparms = Nparms - anasum[2].DF; /* # fitted parameters in fitted model */
  if (anasum[2].DF > 0) {
    if (anasum[2].TEST > .0001) {
      fprintf (fp_out, 
	       "%15s %15.6g %9d %13.6g %6d %15.4g\n"
	       , anatxt[1], anasum[2].SS, Nparms, anasum[2].MSE, anasum[2].DF,
	       anasum[2].TEST);
    }
    else {
      fprintf (fp_out, 
	       "%15s %15.6g %9d %13.6g %6d %19.8g\n"
	       , anatxt[1], anasum[2].SS, Nparms, anasum[2].MSE, anasum[2].DF,
	       anasum[2].TEST);
    }
  }
  else {
    fprintf (fp_out, 
	     "%15s %15.6g %9d %13.6g %6d         NA\n"
	     , anatxt[1], anasum[2].SS, Nparms, anasum[2].MSE, anasum[2].DF);
  }
  Nparms = 1; /* reduced model has just 1 parameter */
  if (anasum[1].DF > 1) {
    if (anasum[3].TEST > .0001) {
      fprintf (fp_out, 
	       "%15s %15.6g %9d %13.6g %6d %15.4g\n"
	       , anatxt[2], anasum[3].SS, Nparms, anasum[3].MSE, anasum[3].DF,
	       anasum[3].TEST);
    }
    else {
      fprintf (fp_out, 
	       "%15s %15.6g %9d %13.6g %6d         <.0001\n"
	       , anatxt[2], anasum[3].SS, Nparms, anasum[3].MSE, anasum[3].DF);
    }
  }
  else {
    fprintf (fp_out, 
	     "%15s %15.6g %9d %13.6g %6d         NA\n"
	     , anatxt[2], anasum[3].SS, Nparms, anasum[3].MSE, anasum[3].DF);
  }

  fprintf (fp_out,
	   "\n%15s %15.6g\n"
	   , "AIC:",
	   -2*anasum[2].SS + 2*(1.0 + anasum[3].DF - anasum[2].DF));

} // myOUTPUT_DTMS3ANOVA()
